from amaranth import Module, Cat, Signal, Assert, unsigned
from amaranth.build import Platform
from amaranth.lib import wiring, data
from amaranth.lib.wiring import In, Out


class RGBLayout(data.StructLayout):
    def __init__(self, r_bits, g_bits, b_bits):
        super().__init__(
            {
                "red": unsigned(r_bits),
                "green": unsigned(g_bits),
                "blue": unsigned(b_bits),
            }
        )


rgb888_layout = RGBLayout(8, 8, 8)

rgb111_hub75 = RGBLayout(1, 1, 1)


class Hub75Stream(wiring.Signature):
    """A Hub75E Driver for a single string of panels."""

    def __init__(self):
        super().__init__(
            {
                "latch": Out(1),
                "oe": Out(1),
                "addr": Out(5),
                "rgb0": Out(3),
                "rgb1": Out(3),
                "display_clk": Out(1),
            }
        )


class Hub75Ctrl(wiring.Signature):
    """Hub75E control/addressing signals.
    This is separated because certain designs keep these common
    between all panel outputs (e.g Colorlight 5A-75b boards).

    """

    def __init__(self):
        super().__init__(
            {
                "latch": Out(1),
                "oe": Out(1),
                "addr": Out(5),
                "display_clk": Out(1),
            }
        )


class Hub75Data(wiring.Signature):
    """ Data lines for HUB75 displays. When combined with Hub75Ctrl, this forms a complete HUB75 interface.
    These are kept separate as some devices have a single set of control signals.
    """
    def __init__(self):
        super().__init__(
            {
                "rgb0": Out(rgb111_hub75),
                "rgb1": Out(rgb111_hub75),
            }
        )
class Hub75StringDriver(wiring.Component):
    """ A data driver for Hub75 panels. This accesses the line memory and feeds out the data
    """
    bcm_select: In(3)


class Hub75Coordinator(wiring.Component):
    """ A shared-control hub75 driver"""
    pass

class Bitslicer(wiring.Component):
    start_write: In(1)
    done: Out(1)
    bitplane_addr: Out(11)
    bitplane_wren: Out(1)
    bitplane_data: Out(6)

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        bitplane_bit = Signal(3)
        with m.FSM():
            with m.State("init"):
                m.d.sync += bitplane_bit.eq(0)
                m.d.sync += self.done.eq(0)

        return m


class Hub75EDriver(wiring.Component):
    """An optimized driver for hub75 strings.
    This version is faster than most implementations by merging the exposure
    period and the data-write period to happen simultaneously. As a result,
    the display can be brighter due to higher duty cycle.
    """

    start: In(1)
    done: Out(1)
    out: Out(Hub75Stream())
    buf_addr: Out(9)
    buf_data: In(36)

    row_depth = 128
    bit_depth = 8
    bcm_len = 32

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(32)
        bcm_shift: Signal = Signal(4, init=7)
        m.d.sync += counter.eq(counter + 1)
        # construct helper signals.
        ram_r = self.buf_data[16:23]
        ram_b = self.buf_data[8:15]
        ram_g = self.buf_data[0:7]

        ram_rgb_slice = Cat(
            ram_r.bit_select(bcm_shift, 1),
            ram_g.bit_select(bcm_shift, 1),
            ram_b.bit_select(bcm_shift, 1),
        )
        pixnum = Signal(8, reset=127)
        pixrow = Signal(1, reset=0)
        m.d.comb += self.buf_addr.eq(Cat(pixrow, pixnum))

        should_clock = counter < (128 * 2 + 1)
        should_expose = (counter < (32 << (bcm_shift + 1) + 1)).bool() & (
            bcm_shift != 7
        ).bool()

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    bcm_shift.eq(7),
                    counter.eq(0),
                    self.done.eq(0),
                    pixnum.eq(127),
                    pixrow.eq(0),
                ]
                with m.If(self.start == 1):
                    m.next = "prefetch"
            with m.State("prefetch"):
                with m.If(counter == 0):
                    m.d.sync += pixrow.eq(0)
                with m.Elif(counter == 1):
                    m.d.sync += pixrow.eq(1)
                with m.Elif(counter == 2):
                    m.d.sync += self.out.rgb0.eq(ram_rgb_slice)
                with m.Elif(counter == 3):
                    m.d.sync += [self.out.rgb1.eq(ram_rgb_slice), counter.eq(0)]
                    m.next = "writerow"
            with m.State("writerow"):
                # expose if we haven't done it for long enough yet.
                m.d.sync += self.out.oe.eq(~(should_expose))
                with m.If(should_clock):
                    # clock is high on entry
                    m.d.sync += self.out.display_clk.eq(counter[1] == 0)

                    with m.If(counter[0:1] == 0):
                        # rising edge of the clock
                        m.d.sync += pixrow.eq(0)

                    with m.If(counter[0:1] == 1):
                        m.d.sync += [
                            pixrow.eq(1),
                            self.out.rgb0.eq(ram_rgb_slice),
                        ]
                    with m.If(counter[0:1] == 2):
                        m.d.sync += [
                            pixnum.eq(pixnum - 1),
                            pixrow.eq(0),
                            self.out.rgb1.eq(ram_rgb_slice),
                        ]

                with m.Elif(~(should_expose)):
                    # we are done both feeding in the new data and exposing the previous.
                    m.d.sync += [counter.eq(0), self.out.display_clk.eq(0)]
                    m.next = "latchout"
            with m.State("latchout"):
                m.d.sync += [
                    pixnum.eq(127),
                    self.out.latch.eq(1),
                ]
                with m.If(counter > 3):
                    m.d.sync += self.out.latch.eq(0)
                    m.d.sync += counter.eq(0)
                    with m.If(bcm_shift == 0):
                        m.next = "finish"
                    with m.Else():
                        m.d.sync += bcm_shift.eq(bcm_shift - 1)
                        m.next = "prefetch"

            with m.State("finish"):
                m.d.sync += Assert(bcm_shift == 0, "finish without bcm shift 0")

                with m.If(counter < (32)):
                    m.d.sync += self.out.oe.eq(0)
                with m.Else():
                    m.d.sync += [self.out.oe.eq(1), self.done.eq(1)]
                    m.next = "init"

        return m
